#include "Definitions.h"
#include "PS2.h"
#include <SoftwareSerial.h>
#include <PS2X_lib.h>  //for v1.6
#include "PS2.h"
#define cSSC_OUT     12         //Output pin for (SSC32 RX) on BotBoard (Yellow)
#define cSSC_IN      13         //Input pin for (SSC32 TX) on BotBoard (Blue)
#define cSSC_BAUD        38400   //SSC32 BAUD rate

SoftwareSerial SSCSerial(cSSC_IN, cSSC_OUT); // where I have the pins defined in a config file:

#define nPhases 10        // Number of phases for this gait
#define nLegs   6        // Number of legs on robot

int time = 250;    // Time (ms) for servos to complete a phase */

float pi = 3.14159;
float muSecPerRad = 2000/pi;  //2000 muSec pulse width = 180 deg (pi radians)

float H = 13.;    // Distance from hip horizontal axis to ground
float d = 8.5;    // Distance from hip vertical axis out to path on ground (Hip at reference position)
float L0 = 3.;    // Distance from hip horizal axis to hip vertical axis
float L1 = 5.7;   // Distance from hip vertical axis to knee axis
float L2 = 13.;   // Distance from hip vertical axis to end of leg
float deg_per_rad = 180/pi;
String command;
int sign;
//float hipUp = 150;        // pulse width for a lifted leg
float liftAngle = 30;    //hip vertical lift angle for lifted legs (deg)
float hipUp = -liftAngle/deg_per_rad;    // Convert liftAngle to radians

int up = 1;
int dn = 0;	// down

int gait_1_positions[nPhases][nLegs] = {
	{dn, up, dn, up, dn, up},
	{dn, up, dn, up, dn, up},
	{dn, up, dn, up, dn, up},
	{dn, up, dn, up, dn, up},
        {dn, dn, dn, dn, dn, dn},
	{up, dn, up, dn, up, dn},
	{up, dn, up, dn, up, dn},
	{up, dn, up, dn, up, dn},
	{up, dn, up, dn, up, dn},
        {dn, dn, dn, dn, dn, dn}
	};
	
float    gait_1_angles[nPhases][nLegs] = {  /* values for alfa for each phase each leg */
	{-30.0, +30.0, -30.0, +30.0, -30.0, +30.0},
	{-16.1, +16.1, -16.1, +16.1, -16.1, +16.1},
	{+0.00, +0.00, +0.00, +0.00, +0.00, +0.00},
	{+16.1, -16.1, +16.1, -16.1, +16.1, -16.1},
	{+30.0, -30.0, +30.0, -30.0, +30.0, -30.0},
	{+30.0, -30.0, +30.0, -30.0, +30.0, -30.0},
	{+16.1, -16.1, +16.1, -16.1, +16.1, -16.1},
	{+0.00, +0.00, +0.00, +0.00, +0.00, +0.00},
	{-16.1, +16.1, -16.1, +16.1, -16.1, +16.1},
	{-30.0, +30.0, -30.0, +30.0, -30.0, +30.0}
	};


leg_servo_inputs leg[nLegs];  // Store IK results in this struct (Definitions.h)

// Actual calibrated servoRefs will be read from the SSC-32 initial-pulse-width registers
int servoRefs[nLegs][3] = {
	{1500, 1500, 1500}, {1500, 1500, 1500}, {1500, 1500, 1500},
	{1500, 1500, 1500}, {1500, 1500, 1500}, {1500, 1500, 1500}};
					
int servo[nLegs][3] = 
	{{0, 1, 2}, {4, 5, 6}, {8, 9, 10}, {16, 17, 18}, {20, 21, 22}, {24, 25, 26}};

void setupPS2();

void InverseKinematics(int phase, int leg, boolean leftSide, float alfa, leg_servo_inputs *legPtr){
	// alfa = hipVertical rotation
        // theta1 = hipVertical rotation
        // theta2 = knee rotation
        // They are zero at the calibrated servo positions
        float D, R, Rsq, theta1, theta2, phi, beta, zeta;
	float tpb;
        int sign = -1;
        if(leftSide){sign = 1;}
        legPtr->hip_H = sign*(int)(alfa*muSecPerRad);
        if(gait_1_positions[phase][leg] == dn){  //Do IK only for down legs
          D = d/cos(alfa);
          Rsq = H*H + (D - L0)*(D - L0);
          R = sqrt(Rsq);
          zeta = acos(H/R);
          beta = pi/2 - zeta;
          // Use lawof cosines to get acos values
          phi = acos( (L1*L1 + L2*L2 - Rsq)/(2*L1*L2) );
          tpb = acos( (L1*L1 + Rsq - L2*L2)/(2*L1*R) );  //theta plus beta
          theta1 = tpb - beta;
          theta2 = pi - phi - pi/2;
          legPtr->hip_V = sign*(int)(theta1*muSecPerRad);
          legPtr->knee = sign*(int)(theta2*muSecPerRad);
          }
        else{
          legPtr->hip_V = sign*(int)(hipUp*muSecPerRad);
          legPtr->knee = sign*(int)(hipUp*muSecPerRad);
          }
}

int ServoPosition(int i, int j, int value){
  return (servoRefs[i][j] + value);
}


String formCommand(int i, int j, int value){
	  command = command + " #";
          command = command + servo[i][j];
          command = command + " P";
          command = command + ServoPosition(i, j, value);
          return command;
}

//==============================================================================
// Quick and dirty helper function to read so many bytes in from the SSC with a timeout and an end of character marker...
//==============================================================================
int SSCRead (byte* pb, int cb, word wTimeout, word wEOL)
{
    int ich;
    byte* pbIn = pb;
    unsigned long ulTimeLastChar = micros();
    while (cb) {
        while (!SSCSerial.available()) {
            // check for timeout
            if ((word)(micros()-ulTimeLastChar) > wTimeout) {
                return (int)(pb-pbIn);
            }    
        }
        ich = SSCSerial.read();
        *pb++ = (byte)ich;
        cb--;

        if ((word)ich == wEOL)
            break;    // we matched so get out of here.
        ulTimeLastChar = micros();    // update to say we received something
    }

    return (int)(pb-pbIn);
}

void FindServoInitialSettings() {  // Read initial pulse widths from SSC-32 
    byte szTemp[5];
    byte cbRead;

    int data;
    short servoNo ; 			// which servo number
    boolean fNew = true;	// is this a new servo to work with?
    boolean fExit = false;	// when to exit
    int ich;
    
    for (int i = 0; i < nLegs; i++) {    // Six legs
      for (int j = 0; j < 3; j++) {  // Three servos per leg
        servoNo = servo[i][j];        // Table of servo numbers
        SSCSerial.print("R");
        SSCSerial.println(64 + servoNo, DEC);  // Register 64, start of initial pulse widths
        // now read in the current value
        cbRead = SSCRead((byte*)szTemp, sizeof(szTemp), 10000, 13);
        if (cbRead > 0){
          servoRefs[i][j] = atoi((const char *)szTemp);
          Serial.print("Servo Ref's [i][j]");
          Serial.print(i);
          Serial.print(" ");
          Serial.print(j);
          Serial.print(" ");
          Serial.println(servoRefs[i][j]);
        }
      else{
        Serial.println("SSCRead failed!");
      }
      }
    }
  }

boolean phaseDone(){ // Check to see if a phase is complete
    char * done;
    byte szTemp[2];
    byte cbRead;
    SSCSerial.println("Q");
    // now read in the current value
    cbRead = SSCRead((byte*)szTemp, sizeof(szTemp), 10000, 13);
    if (cbRead > 0){
      done = (char *)szTemp;
      if(* done == '.'){
        Serial.println("Phase completed OK!");
        return true;
        }
      else{        
        Serial.println("Phase incomplete.");
        return false;
        }
      }
    else{
      Serial.println("SSCRead failed in phaseStatus()!");
      }
    }

void boot(){
  // Then I need to have the SSC-32 reboot in order to use the new values.
  delay(10);    // give it some time to write stuff out.
  SSCSerial.println("GOBOOT");
  delay(5);        // Give it a little time
  SSCSerial.println("g0000");    // tell it that we are done in the boot section so go run the normall SSC stuff...
  delay(500);                // Give it some time to boot up...
  }

void setup(){
  delay(5000);    // 5 sec delay for turning on Serial Monitor
  SSCSerial.begin(38400);  // Enable SSCSerial
  Serial.begin(38400);     // Enable Serial
 // setupPS2();
  for(int phase = 0; phase < nPhases; phase++){
    for(int i = 0; i < nLegs; i++){  /* Convert gait angles to radians */
      gait_1_angles[phase][i] = gait_1_angles[phase][i]/deg_per_rad;
      }
    }
  FindServoInitialSettings();
//  boot();    // Program fails if I use boot()!
  }

void loop(){
  for(int phase = 0; phase < nPhases; phase++){
    command = String("");  /* New command string for each phase */
    for(int legNo = 0; legNo < nLegs; legNo++){
        boolean leftSide;
        int value;
        int hipV;
        if( legNo < 3){  // Right side of robot
          sign = -1;
          hipV = (int)+hipUp;
          leftSide = false;
          } 
        else{        // Left side of robot
          sign = +1;
          hipV = (int)-hipUp;
          leftSide = true;
          }
        InverseKinematics(phase, legNo, leftSide, gait_1_angles[phase][legNo], &leg[legNo]);
        command = formCommand(legNo, 0, leg[legNo].hip_H);
        command = formCommand(legNo, 1, leg[legNo].hip_V);
        command = formCommand(legNo, 2, leg[legNo].knee);
        }   
    command = command + " T";
    command = command + time;
    boolean wait = true;
    while(wait){    // Loop here till previous phase is completed
      if( phaseDone() ){
        wait = false;
        }
      }
    SSCSerial.println(command);  // Send command to the SSC-32
    Serial.print("Phase: ");
    Serial.println(phase);
    Serial.println(command);     // Send to Serial Monitor too for debugging
    }    // End of a phase
}


